"""
europilot.exceptions
~~~~~~~~~~~~~~~~~~~~

Defines custom exceptions

"""


class EuropilotException(Exception):
    pass


class ScreenException(EuropilotException):
    pass


class TrainException(EuropilotException):
    pass
